
/*
  minichoo
  includes: nanobus & nanorouter modified and inline
*/

var assert = {
  ok: function (input, message) {
    if (!input) throw Error((message) ? message : 'An error has occured')
  },
  equal: function (a, b, message) {
    if (a !== b) throw Error((message) ? message : 'An error has occured')
  }
}

var removeItem = function removeItems (arr, startIdx, removeCount) {
  var i, length = arr.length
  if (startIdx >= length || removeCount === 0) {
	return
  }
  removeCount = (startIdx + removeCount > length ? length - startIdx : removeCount)
  var len = length - removeCount
  for (i = startIdx; i < len; ++i) {
	arr[i] = arr[i + removeCount]
  }
  arr.length = len
}

function Nanobus (name) {
  if (!(this instanceof Nanobus)) return new Nanobus(name)
  this._name = name || 'nanobus'
  this._starListeners = []
  this._listeners = {}
}

Nanobus.prototype.emit = function (eventName) {
  assert.ok(typeof eventName === 'string' || typeof eventName === 'symbol', 'nanobus.emit: eventName should be type string or symbol')
  var data = []
  for (var i = 1, len = arguments.length; i < len; i++) {
	data.push(arguments[i])
  }
  var listeners = this._listeners[eventName]
  if (listeners && listeners.length > 0) {
	this._emit(this._listeners[eventName], data)
  }
  if (this._starListeners.length > 0) {
	this._emit(this._starListeners, eventName, data, emitTiming.uuid)
  }
  return this
}

Nanobus.prototype.on = Nanobus.prototype.addListener = function (eventName, listener) {
  assert.ok(typeof eventName === 'string' || typeof eventName === 'symbol', 'nanobus.on: eventName should be type string or symbol')
  assert.equal(typeof listener, 'function', 'nanobus.on: listener should be type function')
  if (eventName === '*') {
	this._starListeners.push(listener)
  } else {
	if (!this._listeners[eventName]) this._listeners[eventName] = []
	this._listeners[eventName].push(listener)
  }
  return this
}

Nanobus.prototype.prependListener = function (eventName, listener) {
  assert.ok(typeof eventName === 'string' || typeof eventName === 'symbol', 'nanobus.prependListener: eventName should be type string or symbol')
  assert.equal(typeof listener, 'function', 'nanobus.prependListener: listener should be type function')
  if (eventName === '*') {
	this._starListeners.unshift(listener)
  } else {
	if (!this._listeners[eventName]) this._listeners[eventName] = []
	this._listeners[eventName].unshift(listener)
  }
  return this
}

Nanobus.prototype.once = function (eventName, listener) {
  assert.ok(typeof eventName === 'string' || typeof eventName === 'symbol', 'nanobus.once: eventName should be type string or symbol')
  assert.equal(typeof listener, 'function', 'nanobus.once: listener should be type function')
  var self = this
  this.on(eventName, once)
  function once () {
	listener.apply(self, arguments)
	self.removeListener(eventName, once)
  }
  return this
}

Nanobus.prototype.prependOnceListener = function (eventName, listener) {
  assert.ok(typeof eventName === 'string' || typeof eventName === 'symbol', 'nanobus.prependOnceListener: eventName should be type string or symbol')
  assert.equal(typeof listener, 'function', 'nanobus.prependOnceListener: listener should be type function')
  var self = this
  this.prependListener(eventName, once)
  function once () {
	listener.apply(self, arguments)
	self.removeListener(eventName, once)
  }
  return this
}

Nanobus.prototype.removeListener = function (eventName, listener) {
  assert.ok(typeof eventName === 'string' || typeof eventName === 'symbol', 'nanobus.removeListener: eventName should be type string or symbol')
  assert.equal(typeof listener, 'function', 'nanobus.removeListener: listener should be type function')
  if (eventName === '*') {
	this._starListeners = this._starListeners.slice()
	return remove(this._starListeners, listener)
  } else {
	if (typeof this._listeners[eventName] !== 'undefined') {
	  this._listeners[eventName] = this._listeners[eventName].slice()
	}
	return remove(this._listeners[eventName], listener)
  }

  function remove (arr, listener) {
	if (!arr) return
	var index = arr.indexOf(listener)
	if (index !== -1) {
	  splice(arr, index, 1)
	  return true
	}
  }
}

Nanobus.prototype.removeAllListeners = function (eventName) {
  if (eventName) {
	if (eventName === '*') {
	  this._starListeners = []
	} else {
	  this._listeners[eventName] = []
	}
  } else {
	this._starListeners = []
	this._listeners = {}
  }
  return this
}

Nanobus.prototype.listeners = function (eventName) {
  var listeners = eventName !== '*'
	? this._listeners[eventName]
	: this._starListeners
  var ret = []
  if (listeners) {
	var ilength = listeners.length
	for (var i = 0; i < ilength; i++) ret.push(listeners[i])
  }
  return ret
}

Nanobus.prototype._emit = function (arr, eventName, data, uuid) {
  if (typeof arr === 'undefined') return
  if (arr.length === 0) return
  if (data === undefined) {
	data = eventName
	eventName = null
  }
  if (eventName) {
	if (uuid !== undefined) {
	  data = [eventName].concat(data, uuid)
	} else {
	  data = [eventName].concat(data)
	}
  }

  var length = arr.length
  for (var i = 0; i < length; i++) {
	var listener = arr[i]
	listener.apply(listener, data)
  }
}

function has (object, property) {
  return Object.prototype.hasOwnProperty.call(object, property)
}

function Trie () {
  if (!(this instanceof Trie)) return new Trie()
  this.trie = { nodes: {} }
}

Trie.prototype.create = function (route) {
  assert.equal(typeof route, 'string', 'route should be a string')
  var routes = route.replace(/^\//, '').split('/')
  function createNode (index, trie) {
    var thisRoute = (has(routes, index) && routes[index])
    if (thisRoute === false) return trie
    var node = null
    if (/^:|^\*/.test(thisRoute)) {
      if (!has(trie.nodes, '$$')) {
        node = { nodes: {} }
        trie.nodes.$$ = node
      } else {
        node = trie.nodes.$$
      }
      if (thisRoute[0] === '*') {
        trie.wildcard = true
      }
      trie.name = thisRoute.replace(/^:|^\*/, '')
    } else if (!has(trie.nodes, thisRoute)) {
      node = { nodes: {} }
      trie.nodes[thisRoute] = node
    } else {
      node = trie.nodes[thisRoute]
    }
    return createNode(index + 1, node)
  }
  return createNode(0, this.trie)
}

Trie.prototype.match = function (route) {
  assert.equal(typeof route, 'string', 'route should be a string')
  var routes = route.replace(/^\//, '').split('/')
  var params = {}
  function search (index, trie) {
    if (trie === undefined) return undefined
    var thisRoute = routes[index]
    if (thisRoute === undefined) return trie
    if (has(trie.nodes, thisRoute)) {
      return search(index + 1, trie.nodes[thisRoute])
    } else if (trie.name) {
      try {
        params[trie.name] = decodeURIComponent(thisRoute)
      } catch (e) {
        return search(index, undefined)
      }
      return search(index + 1, trie.nodes.$$)
    } else if (trie.wildcard) {
      try {
        params.wildcard = decodeURIComponent(routes.slice(index).join('/'))
      } catch (e) {
        return search(index, undefined)
      }
      return trie.nodes.$$
    } else {
      return search(index + 1)
    }
  }
  var node = search(0, this.trie)
  if (!node) return undefined
  node = Object.assign({}, node)
  node.params = params
  return node
}

Trie.prototype.mount = function (route, trie) {
  assert.equal(typeof route, 'string', 'route should be a string')
  assert.equal(typeof trie, 'object', 'trie should be a object')
  var split = route.replace(/^\//, '').split('/')
  var node = null
  var key = null
  if (split.length === 1) {
    key = split[0]
    node = this.create(key)
  } else {
    var head = split.join('/')
    key = split[0]
    node = this.create(head)
  }
  Object.assign(node.nodes, trie.nodes)
  if (trie.name) node.name = trie.name
  if (node.nodes['']) {
    Object.keys(node.nodes['']).forEach(function (key) {
      if (key === 'nodes') return
      node[key] = node.nodes[''][key]
    })
    Object.assign(node.nodes, node.nodes[''].nodes)
    delete node.nodes[''].nodes
  }
}

function Wayfarer (dft) {
  if (!(this instanceof Wayfarer)) return new Wayfarer(dft)
  var _default = (dft || '').replace(/^\//, '')
  var _trie = Trie()
  emit._trie = _trie
  emit.on = on
  emit.emit = emit
  emit.match = match
  emit._wayfarer = true
  return emit
  function on (route, cb) {
    assert.equal(typeof route, 'string')
    assert.equal(typeof cb, 'function')
    route = route || '/'
    if (cb._wayfarer && cb._trie) {
      _trie.mount(route, cb._trie.trie)
    } else {
      var node = _trie.create(route)
      node.cb = cb
      node.route = route
    }
    return emit
  }
  function emit (route) {
    var matched = match(route)
    var args = new Array(arguments.length)
    args[0] = matched.params
    for (var i = 1; i < args.length; i++) {
      args[i] = arguments[i]
    }
    return matched.cb.apply(matched.cb, args)
  }
  function match (route) {
    if (!route) throw Error("'route' must be defined")
    var matched = _trie.match(route)
    if (matched && matched.cb) return new Route(matched)
    var dft = _trie.match(_default)
    if (dft && dft.cb) return new Route(dft)
    throw new Error("route '" + route + "' did not match")
  }
  function Route (matched) {
    this.cb = matched.cb
    this.route = matched.route
    this.params = matched.params
  }
}

var protocol = '^(http(s)?(:\/\/))?(www\.)?'
var domain = '[a-zA-Z0-9-_\.]+(:[0-9]{1,5})?(\/{1})?'
var qs = '[\?].*$'
var prefix = new RegExp(protocol + domain)
var normalize = new RegExp('#')
var suffix = new RegExp(qs)

function Nanorouter (opts) {
  if (!(this instanceof Nanorouter)) return new Nanorouter(opts)
  opts = opts || {}
  this.router = Wayfarer(opts.default || '/404')
}

Nanorouter.prototype.on = function (routename, listener) {
  assert.equal(typeof routename, 'string')
  routename = routename.replace(/^[#/]/, '')
  this.router.on(routename, listener)
}

Nanorouter.prototype.emit = function (routename) {
  assert.equal(typeof routename, 'string')
  routename = pathname(routename)
  return this.router.emit(routename)
}

Nanorouter.prototype.match = function (routename) {
  assert.equal(typeof routename, 'string')
  routename = pathname(routename)
  return this.router.match(routename)
}

function pathname (routename) {
  routename = routename.replace(prefix, '')
  return decodeURI(routename.replace(suffix, '').replace(normalize, '/'))
}

// minichoo
//
function minichoo () {
  if (! (this instanceof minichoo)) return new minichoo ()
  this.middleware = []
  this.routes = {}
  this.state = {}
  this.params = {}
  var self = this
  this.emitter = Nanobus()
  this.router = Nanorouter()
  setTimeout(() => {
    this.app = document.getElementById("app")
  })
  window.onhashchange = function() {
    var current_route = location.hash.replace(/\#/g, '') || '/#';
    // <a href="#!" will be ignored...
    if (!current_route.includes('!')) self.router.emit(current_route)
  }
}

minichoo.prototype._finish_setup = function () {
  var self = this
  Object.keys(self.routes).forEach(function (route) {
    self.router.on(route, function (params) {
      self.state.params = params
      self.app.innerHTML = self.routes[route](self.state, self.emitter)
    })
  })
}

minichoo.prototype.use = function (middleware) {
  assert.ok(typeof middleware === 'function')
  this.middleware.push(middleware)
}

minichoo.prototype.route = function (route, handler) {
  assert.ok(typeof route === 'string')
  assert.ok(typeof handler === 'function')
  this.routes[route] = handler;
}

minichoo.prototype.mount = function (element) {
  var self = this;
  var default_middleware = function (state, emit) {
    state.ready = function (...args) { return setTimeout(...args); }
    state.form = function addForm (id, callback) {
      var form = document.getElementById(id)
      form.addEventListener('submit', function (e) {
        e.preventDefault()
        var data = new FormData(e.currentTarget)
        callback(data, e)
      })
    }
  }
  default_middleware(self.state, self.emitter)
  self.middleware.forEach(function (middleware) {
    middleware(self.state, self.emitter)
  })
  self._finish_setup()
  setTimeout(function () {
    self.app.innerHTML = self.routes[
    Object.keys(self.routes)[0]
    ](self.state, self.emitter)
  }, 1000)
}

window.minichoo = minichoo;
